@@ -135,8 +135,11 @@ void nt_print_builtins(unsigned int maxwidth) {
 }
 void dostart(Char ** vc, struct command *c) {
 
-	char cmdstr[512];
+	char *cmdstr,*cmdend,*ptr;
+	char argv0[256];
+	DWORD cmdsize;
 	char *currdir=NULL;
+	char *savepath;
 	char **v;
 	STARTUPINFO si;
 	PROCESS_INFORMATION pi;
@@ -146,8 +149,11 @@ void dostart(Char ** vc, struct command *c) {
 
 	USE(c);*vc++;
 
-	cmdlen = 512;
-	memset(cmdstr,0,512);
+	cmdsize = 512;
+	cmdstr = heap_alloc(cmdsize);
+	cmdend = cmdstr;
+	cmdlen = 0;
+
 	memset(&si,0,sizeof(si));
 	si.cb = sizeof(si);
 
@@ -201,38 +207,44 @@ void dostart(Char ** vc, struct command *c) {
 	 * -amol 5/30/96
 	 */
 	for (jj=k;v[jj] != NULL; jj++) {
-			j=(strlen(v[jj]) + 2);
-			if(j > cmdlen) {
-				blkfree((Char **)v);
-				stderror(ERR_TOOMANY);
+			j=(lstrlen(v[jj]) + 2);
+			if (j + cmdlen > cmdsize) {
+				ptr = cmdstr;
+				cmdstr = heap_realloc(cmdstr, cmdsize << 1);
+				cmdend =  cmdstr + (cmdend - ptr);
+				cmdsize <<= 1;
 			}
-			lstrcat(cmdstr,v[jj]);
-			lstrcat(cmdstr," ");
+			ptr = v[jj];
+			while (*ptr) {
+				*cmdend++ = *ptr++;
+				cmdlen++;
+			}
+			*cmdend++ = ' ';
+			cmdlen++;
 	}
 	if (jj == k) {
 		blkfree((Char **)v);
 		stderror(ERR_SYSTEM,start_usage,"See CMD.EXE for more info");
 	}
+	*cmdend = 0;
+	lstrcpyn(argv0,v[k],255);
+
 	{
-		char pathstr[1024];
-		char *ptr;
-		Char *vp;
+		DWORD len;
 
-		vp = varval(STRNTlamepathfix);
+		/* 
+		 * strictly speaking, it should do no harm to set the path
+		 * back to '\'-delimited even in the parent, but in the
+		 * interest of consistency, we save the old value and restore it
+		 * later
+		 */
+		len = GetEnvironmentVariable("PATH",NULL,0);
+		savepath = heap_alloc(len+1);
+		GetEnvironmentVariable("PATH",savepath,len);
+
+		fix_path_for_child();
+	}
 
-		pathstr[1023] =0 ;
-		if (vp != STRNULL) {
-			if (GetEnvironmentVariable("PATH",pathstr,1024) <1024 ) {
-				ptr = pathstr;
-				while(*ptr) {
-					if (*ptr == '/')
-						*ptr = '\\';
-					ptr++;
-				}
-				SetEnvironmentVariable("PATH",pathstr);
-			}
-		}
-	}			
 	if (! CreateProcess(NULL,
 						cmdstr,
 						NULL,
@@ -243,33 +255,59 @@ void dostart(Char ** vc, struct command *c) {
 						currdir,
 						&si,
 						&pi) ) {
+
+		SetEnvironmentVariable("PATH",savepath);
+		heap_free(savepath);
+
 		ret = GetLastError();
 		if (ret == ERROR_BAD_EXE_FORMAT || ret == ERROR_ACCESS_DENIED ||
 				(ret == ERROR_FILE_NOT_FOUND && is_url(v[k])) ) {
+
 			errno = ENOEXEC;
+
 			if (!__nt_only_start_exes)
 				try_shell_ex(&v[k],0);
-			if (errno)
-				stderror(ERR_ARCH,cmdstr,strerror(errno));
+
+			heap_free(cmdstr); /* free !! */
+
+			if (errno) {
+				stderror(ERR_ARCH,argv0,strerror(errno));
+			}
 		}
-		else{
+		else if (ret == ERROR_INVALID_PARAMETER) {
+
+			errno = ENAMETOOLONG;
+
+			heap_free(cmdstr); /* free !! */
+
+			stderror(ERR_TOOLARGE,argv0);
+
+		}
+		else {
 			errno = ENOENT;
 			if (
-				( (v[k][0] == '\\') ||(v[k][0] == '/') ) &&
-				( (v[k][1] == '\\') ||(v[k][1] == '/') ) &&
-				(!v[k+1])
+					( (v[k][0] == '\\') ||(v[k][0] == '/') ) &&
+					( (v[k][1] == '\\') ||(v[k][1] == '/') ) &&
+					(!v[k+1])
 			   )
 				try_shell_ex(&v[k],0);
-			if (errno)
-				stderror(ERR_NOTFOUND,cmdstr);
+
+			heap_free(cmdstr); /* free !! */
+			if (errno) {
+				stderror(ERR_NOTFOUND,argv0);
+			}
 		}
 	}
-	CloseHandle(pi.hProcess);
-	CloseHandle(pi.hThread);
+	else {
+		CloseHandle(pi.hProcess);
+		CloseHandle(pi.hThread);
 
+		SetEnvironmentVariable("PATH",savepath);
+		heap_free(cmdstr);
+		heap_free(savepath);
+	}
 	blkfree((Char **)v);
 	return;
-
 }
 void error(char * ebuf) {
 
@@ -311,7 +349,7 @@ void nt_set_env(Char *name, Char *val) {
 		if (!nameBuf) {
 			stderror(ERR_TOOLARGE);
 		}
-		strcpy(nameBuf,cname);
+		lstrcpy(nameBuf,cname);
 	}
 	cval = val?short2str(val):NULL;
 	if(cval) {
@@ -321,7 +359,7 @@ void nt_set_env(Char *name, Char *val) {
 		if (len >= BUFSIZ) 
 			stderror(ERR_TOOLARGE);
 		*/
-		strcpy(valBuf,cval);
+		lstrcpy(valBuf,cval);
 	}
 	
 	SetEnvironmentVariable(nameBuf,cval?valBuf:NULL);
@@ -457,6 +495,7 @@ int nt_feed_to_cmd(char *file,char **argv) {
 	CloseHandle(pi.hProcess);
 	ExitProcess(0);
 
+	return 1; /*NOTREACHED*/
 }
 static char *hb_subst_array[20] ;
 void init_hb_subst(void) {
@@ -625,15 +664,11 @@ void try_shell_ex(char **argv,int exitsuccess) {
 
 	}
 	cmdstr[0] = 0;
-	if ( (!*argv) &&  (prog[0] == '\\') && (prog[1] == '\\')) {
-		mask |= SEE_MASK_CONNECTNETDRV;
-		nocmd = 1;
-		goto noargs;
-	}
 	while(*argv) {
 
 		ptr = *argv;
 
+		lstrcat(cmdstr," ");
 		while(*ptr) {
 			if (*ptr == ' ')
 			quotespace=1;
@@ -648,8 +683,6 @@ void try_shell_ex(char **argv,int exitsuccess) {
 			lstrcat(cmdstr,*argv);
 		*argv++;
 	}
-	//dprintf("executing %s %s\n",prog,cmdstr);
-noargs:
 
 	memset(&shinfo,0,sizeof(shinfo));
 	shinfo.cbSize = sizeof(shinfo);
@@ -657,7 +690,7 @@ noargs:
 	shinfo.hwnd = NULL;
 	shinfo.lpVerb = NULL;
 	shinfo.lpFile = prog;
-	shinfo.lpParameters = nocmd?NULL:&cmdstr[0];
+	shinfo.lpParameters = &cmdstr[0];
 	shinfo.lpDirectory = 0;
 	shinfo.nShow = SW_SHOWDEFAULT;
 
@@ -745,7 +778,7 @@ int nt_try_fast_exec(struct command *t) {
 	xfree((ptr_t) pv);
 	xfree((ptr_t) av);
 	av = t->t_dcom;
-	trim(av);
+	//trim(av);
 
 	if (*av == NULL || **av == '\0')
 		return 1;
@@ -804,84 +837,47 @@ int nt_texec(char *prog, char**args ) {
 	DWORD dwCreationflags;
 	unsigned int priority;
 	char *argv0;
-	char *cmdstr ;
-	unsigned int cmdsize,cmdlen,arglen;
-    char *ptr,*p2;//,*p3;
+	char *cmdstr,*cmdend ,*cmdstr_save;
+	unsigned int cmdsize,cmdlen;
+    char *p2;
 	char **savedargs;
 	int retries=0;
 	int hasdot =0;
 	int is_winnt, hasspace=0;
 	int retval = 1;
-	char tempquotedbuf[256];
 
 	memset(&si,0,sizeof(si));
 	savedargs = args;
 
 	/* MUST FREE !! */
-	cmdstr= heap_alloc(MAX_PATH*4);
-	argv0= heap_alloc(MAX_PATH);
+	cmdstr= heap_alloc(MAX_PATH<<2);
+	cmdstr_save = cmdstr;
+	cmdsize = MAX_PATH<<2;
 
 	is_winnt = (gdwPlatform != VER_PLATFORM_WIN32_WINDOWS);
 
-	ptr = prog;
-	p2 = cmdstr+1; // Leave room for quote if needed
-	//p3 = argv0;
 
-	cmdsize = MAX_PATH*4;
-	cmdlen = 1;
+	p2 = cmdstr;
 
-	/*
-	 * Change '/' to '\\', quote spaces
-	 */
-	while(*ptr) {
-		if (*ptr == '/')
-			*ptr = '\\';
-		else if (*ptr == '.')
-			hasdot = 1;
-		else if (*ptr == ' ' ) {
-			if(ptr[-1] == '\\')
-				p2--;
-			hasspace = 1;
-		}
-		*p2++ = *ptr;
-		//*p3++ = *ptr;
+	cmdlen = 0;
+	cmdlen += copy_quote_and_fix_slashes(prog,cmdstr,&hasdot);
+	p2 += cmdlen;
 
-		ptr++;
-		cmdlen++;
-	}
-
-	//
-	// The fix below is for paths with ./ instead of . on win95
-	//
-	if ( !is_winnt && cmdstr[1] == '.' && cmdstr[2] == '\\' 
-			&& cmdstr[3] == '\\' ) {
-		ptr = cmdstr + 3; // move to 2nd '\';
-		while(* (ptr +1) ){
-			*ptr = *(ptr +1);
-			ptr++;
-		}
-		*ptr = 0;
-
-	}
-	if (hasspace) {
-		/* put quotes at beginning and end */
-		*cmdstr = '"';
-		*p2++ = '"';
-
-		cmdlen++;
-	}
-	else { // Skip initial character we left for quote
+	if (*cmdstr != '"') {
+		 // If not quoted, skip initial character we left for quote
 		*cmdstr = 'A';
 		cmdstr++; 
 	}
-	*p2 = 0; //*p3 = 0;
+	*p2 = 0; 
+	cmdend = p2;
 
 	if (!is_winnt) {
-		//heap_free(argv0);
 		argv0 = NULL;
 	}
-	else
+	else {
+		argv0= heap_alloc(MAX_PATH);
 		wsprintf(argv0,"%s",prog);
+	}
 
 	si.cb = sizeof(STARTUPINFO);
 	si.dwFlags = STARTF_USESTDHANDLES;
@@ -904,66 +900,13 @@ int nt_texec(char *prog, char**args ) {
 
 	*args++; // the first arg is the command
 
-	while(*args) {
-		short quotespace=0;
-		short quotequote=0;
+	concat_args_and_quote(args,&cmdstr,&cmdlen,&cmdend,&cmdsize);
 
-		lstrcat(cmdstr," ");
-		cmdlen++;
+	*cmdend = 0;
 
-		ptr = *args;
-
-		arglen = 0;
-
-		if (!*ptr) {
-			lstrcat(cmdstr,"\"\"");
-			cmdlen += 2;
-		}
-		while(*ptr) {
-			if ((*ptr == ' ' || *ptr == '\t') && (ptr[-1] != '\\'))
-				quotespace=1;
-			else if (*ptr == '"')
-				quotequote = 1;
-			ptr++;
-			arglen++;
-		}
-		if (arglen + cmdlen +4 > cmdsize) { // +4 is if we have to quote
-
-			if (arglen + 8 > cmdsize){
-				cmdstr = heap_realloc(cmdstr-1,cmdsize+arglen+8);
-				cmdstr++;
-				cmdsize += arglen+8;
-			}
-			else{
-				cmdstr = heap_realloc(cmdstr-1,cmdsize<<1);
-				cmdstr++;
-				cmdsize <<=1;
-			}
-
-		}
-		if (quotespace) 
-			lstrcat(cmdstr,"\"");
-
-		if (quotequote){
-			tempquotedbuf[0]=0;
-			quoteProtect(tempquotedbuf,*args);
-			lstrcat(cmdstr,tempquotedbuf);
-			cmdlen +=2;
-		}
-		else
-			lstrcat(cmdstr,*args);
-
-		if (quotespace) {
-			lstrcat(cmdstr,"\"");
-			cmdlen +=2;
-		}
-
-		cmdlen += arglen;
-
-		args++;
-	}
 	dwCreationflags = GetPriorityClass(GetCurrentProcess());
 	priority = GetThreadPriority(GetCurrentThread());
+
 	if (is_winnt) {
 		retries = 0;
 		do {
@@ -993,8 +936,11 @@ re_cp:
 		if (GetLastError() == ERROR_BAD_EXE_FORMAT) {
 			errno  = ENOEXEC;
 		}
-		else
+		else if (GetLastError() == ERROR_INVALID_PARAMETER) {
+			errno = ENAMETOOLONG;
+		}else {
 			errno  = ENOENT;
+		}
 		if (!is_winnt && !hasdot) { //append '.' to the end if needed
 			lstrcat(cmdstr,".");
 			hasdot=1;
@@ -1018,7 +964,9 @@ re_cp:
 		retval = 0;
 	}
 cleanup:
-	heap_free(cmdstr-1);
+	heap_free(cmdstr_save);
+	if (argv0)
+		heap_free(argv0);
 	CloseHandle(si.hStdInput);
 	CloseHandle(si.hStdOutput);
 	CloseHandle(si.hStdError);
